/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

void alloc_free_energy(void)
{
  free_energy = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*free_energy_size);
}

void alloc_free_energy_best(void)
{
  free_energy_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*free_energy_size);
}

void alloc_log_itemp_vb(void)
{
  log_itemp = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*free_energy_size);
  log_itemp_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*free_energy_size);
}

void expand_free_energy_all(void)
{
  int old_free_energy_size;

  old_free_energy_size = free_energy_size;
  free_energy_size *= 2;

  free_energy = (FLOAT_T *)n_realloc(free_energy,sizeof(FLOAT_T)*free_energy_size);
  free_energy_best = (FLOAT_T *)n_realloc(free_energy_best,sizeof(FLOAT_T)*free_energy_size);

  if (mode_daem) {
    log_itemp = (FLOAT_T *)n_realloc(log_itemp,sizeof(FLOAT_T)*free_energy_size);
    log_itemp_best = (FLOAT_T *)n_realloc(log_itemp_best,sizeof(FLOAT_T)*free_energy_size);
  }
}

void alloc_seed_vb(void)
{
  int u,num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  log_final_free_energy = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_search);

  for (u = 0; u < num_search; u++)
    log_final_free_energy[u] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_init);
}

void alloc_hparam_best(void)
{
  int k,v,u,n;

  hparam_c_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);
  hparam_c_upd_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    hparam_w_best = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    hparam_w_upd_best = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      hparam_w_best[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      hparam_w_upd_best[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
	hparam_w_best[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
	hparam_w_upd_best[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
      }
    }
  }

  if (mode_gauss) {
    hparam_u_best = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    hparam_s_best = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p_best = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet_best = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

    hparam_u_upd_best = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    hparam_s_upd_best = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p_upd_best = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet_upd_best = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

    expect_c_best = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

    for (k=0; k<num_class; k++) {
      hparam_u_best[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      hparam_s_best[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_best[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_best[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);

      hparam_u_upd_best[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      hparam_s_upd_best[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_upd_best[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_upd_best[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];

	hparam_u_best[k][u] = alloc_vector(n);
	hparam_s_best[k][u] = alloc_matrix(n,n);
	hparam_p_best[k][u] = alloc_matrix(n,n);

	hparam_u_upd_best[k][u] = alloc_vector(n);
	hparam_s_upd_best[k][u] = alloc_matrix(n,n);
	hparam_p_upd_best[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void save_hparam_best(void)
{
  int k,v,j,u;

  for (k=0; k<num_class; k++) {
    hparam_c_best[k] = hparam_c[k];
    hparam_c_upd_best[k] = hparam_c_upd[k];
  }

  if (mode_nominal) {
#pragma omp parallel for private(v,j)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	for (j=0; j<num_word[v]; j++) {
	  hparam_w_best[k][v][j] = hparam_w[k][v][j];
	  hparam_w_upd_best[k][v][j] = hparam_w_upd[k][v][j];
	}
      }
    }
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(hparam_u_best[k][u],hparam_u[k][u]);
	copy_matrix(hparam_s_best[k][u],hparam_s[k][u]);
	copy_matrix(hparam_p_best[k][u],hparam_p[k][u]);
	hparam_sdet_best[k][u] = hparam_sdet[k][u];

	copy_vector(hparam_u_upd_best[k][u],hparam_u_upd[k][u]);
	copy_matrix(hparam_s_upd_best[k][u],hparam_s_upd[k][u]);
	copy_matrix(hparam_p_upd_best[k][u],hparam_p_upd[k][u]);
	hparam_sdet_upd_best[k][u] = hparam_sdet_upd[k][u];
      }
      expect_c_best[k] = expect_c[k];
    }
  }
}

void restore_hparam_best(void)
{
  int k,v,j,u;

  for (k=0; k<num_class; k++) {
    hparam_c[k] = hparam_c_best[k];
    hparam_c_upd[k] = hparam_c_upd_best[k];
  }

  if (mode_nominal) {
#pragma omp parallel for private(v,j)
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	for (j=0; j<num_word[v]; j++) {
	  hparam_w[k][v][j] = hparam_w_best[k][v][j];
	  hparam_w_upd[k][v][j] = hparam_w_upd_best[k][v][j];
	}
      }
    }
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(hparam_u[k][u],hparam_u_best[k][u]);
	copy_matrix(hparam_s[k][u],hparam_s_best[k][u]);
	copy_matrix(hparam_p[k][u],hparam_p_best[k][u]);
	hparam_sdet[k][u] = hparam_sdet_best[k][u];

	copy_vector(hparam_u_upd[k][u],hparam_u_upd_best[k][u]);
	copy_matrix(hparam_s_upd[k][u],hparam_s_upd_best[k][u]);
	copy_matrix(hparam_p_upd[k][u],hparam_p_upd_best[k][u]);
	hparam_sdet_upd[k][u] = hparam_sdet_upd_best[k][u];
      }
      expect_c[k] = expect_c_best[k];
    }
  }
}

void alloc_pi(void)
{
  int k,v;

  pi_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    pi_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      pi_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	pi_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    pi_g = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
    
    for (k=0; k<num_class; k++) {
      pi_g[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
    }
  }
}

void release_free_energy(void)
{
  free(free_energy); free_energy = NULL;
}

void release_free_energy_best(void)
{
  free(free_energy_best); free_energy_best = NULL;
}

void release_log_itemp_vb(void)
{
  release_log_itemp();
}

void release_seed_vb(void)
{
  int u, num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  for (u=0; u < num_search; u++) {
    free(log_final_free_energy[u]); log_final_free_energy[u] = NULL;
  }
  free(log_final_free_energy); log_final_free_energy = NULL;
}

void alloc_hparam_upd(void)
{
  int k,v,u,n;

  hparam_c_upd = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    hparam_w_upd = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      hparam_w_upd[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
	hparam_w_upd[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
    }
  }

  if (mode_gauss) {
    hparam_u_upd = (VECTOR **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_s_upd = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p_upd = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet_upd = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);
    
    for (k=0; k<num_class; k++) {
      hparam_u_upd[k] = (VECTOR *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_s_upd[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_upd[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_upd[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];

	hparam_u_upd[k][u] = alloc_vector(n);
	hparam_s_upd[k][u] = alloc_matrix(n,n);
	hparam_p_upd[k][u] = alloc_matrix(n,n);
	
	clear_vector(hparam_u_upd[k][u]);
	clear_matrix(hparam_s_upd[k][u]);
	clear_matrix(hparam_p_upd[k][u]);
      }
    }
  }
}

void release_hparam_upd(void)
{
  int k,v,u,n;

  free(hparam_c_upd); hparam_c_upd = NULL;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(hparam_w_upd[k][v]); hparam_w_upd[k][v] = NULL;
      }
      free(hparam_w_upd[k]); hparam_w_upd[k] = NULL;
    }
    free(hparam_w_upd); hparam_w_upd = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];
	release_vector(hparam_u_upd[k][u]);
	release_matrix(hparam_s_upd[k][u]);
	release_matrix(hparam_p_upd[k][u]);
      }
      free(hparam_u_upd[k]);
      free(hparam_s_upd[k]);
      free(hparam_p_upd[k]);
      free(hparam_sdet_upd[k]);
    }
    free(hparam_u_upd);  hparam_u_upd = NULL;
    free(hparam_s_upd);  hparam_s_upd = NULL;
    free(hparam_p_upd);  hparam_p_upd = NULL;
    free(hparam_sdet_upd);  hparam_sdet_upd = NULL;
  }
}

void release_hparam_best(void)
{
  int k,v,u;

  free(hparam_c_best);     hparam_c_best = NULL;
  free(hparam_c_upd_best); hparam_c_upd_best = NULL;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(hparam_w_best[k][v]);     hparam_w_best[k][v] = NULL;
	free(hparam_w_upd_best[k][v]); hparam_w_upd_best[k][v] = NULL;
      }
      free(hparam_w_best[k]);     hparam_w_best[k] = NULL;
      free(hparam_w_upd_best[k]); hparam_w_upd_best[k] = NULL;
    }
    free(hparam_w_best);     hparam_w_best = NULL;
    free(hparam_w_upd_best); hparam_w_upd_best = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(hparam_u_best[k][u]);
	release_matrix(hparam_s_best[k][u]);
	release_matrix(hparam_p_best[k][u]);

	release_vector(hparam_u_upd_best[k][u]);
	release_matrix(hparam_s_upd_best[k][u]);
	release_matrix(hparam_p_upd_best[k][u]);
      }

      free(hparam_u_best[k]);    hparam_u_best[k] = NULL;
      free(hparam_s_best[k]);    hparam_s_best[k] = NULL;
      free(hparam_p_best[k]);    hparam_p_best[k] = NULL;
      free(hparam_sdet_best[k]); hparam_sdet_best[k] = NULL;

      free(hparam_u_upd_best[k]);    hparam_u_upd_best[k] = NULL;
      free(hparam_s_upd_best[k]);    hparam_s_upd_best[k] = NULL;
      free(hparam_p_upd_best[k]);    hparam_p_upd_best[k] = NULL;
      free(hparam_sdet_upd_best[k]); hparam_sdet_upd_best[k] = NULL;
    }

    free(hparam_u_best);    hparam_u_best = NULL;
    free(hparam_s_best);    hparam_s_best = NULL;
    free(hparam_p_best);    hparam_p_best = NULL;
    free(hparam_sdet_best); hparam_sdet_best = NULL;

    free(hparam_u_upd_best);    hparam_u_upd_best = NULL;
    free(hparam_s_upd_best);    hparam_s_upd_best = NULL;
    free(hparam_p_upd_best);    hparam_p_upd_best = NULL;
    free(hparam_sdet_upd_best); hparam_sdet_upd_best = NULL;

    free(expect_c_best);  expect_c_best = NULL;
  }

}

void save_hparam_opt(void)
{
  int k,v,i,u,n;

  hparam_c_opt = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);
  hparam_c_upd_opt = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    hparam_w_opt = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    hparam_w_upd_opt = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    
    for (k=0; k<num_class; k++) {
      hparam_w_opt[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      hparam_w_upd_opt[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
	hparam_w_opt[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
	hparam_w_upd_opt[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
      }
    }
  }

  if (mode_gauss) {
    hparam_u_upd_opt = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    hparam_s_upd_opt = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p_upd_opt = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet_upd_opt = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

    expect_c_opt = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

    for (k=0; k<num_class; k++) {
      hparam_u_upd_opt[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      hparam_s_upd_opt[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_upd_opt[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_upd_opt[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);

      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];

	hparam_u_upd_opt[k][u] = alloc_vector(n);
	hparam_s_upd_opt[k][u] = alloc_matrix(n,n);
	hparam_p_upd_opt[k][u] = alloc_matrix(n,n);
      }
    }
  }

  for (k=0; k<num_class_opt; k++) {
    hparam_c_opt[k] = hparam_c[k];
    hparam_c_upd_opt[k] = hparam_c_upd[k];
  }

  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class_opt; k++) {
      for (v=0; v<tuple_size_d; v++) {
	for (i=0; i<num_word[v]; i++) {
	  hparam_w_opt[k][v][i] = hparam_w[k][v][i];
	  hparam_w_upd_opt[k][v][i] = hparam_w_upd[k][v][i];
	}
      }
    }
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(hparam_u_upd_opt[k][u],hparam_u_upd[k][u]);
	copy_matrix(hparam_s_upd_opt[k][u],hparam_s_upd[k][u]);
	copy_matrix(hparam_p_upd_opt[k][u],hparam_p_upd[k][u]);
	hparam_sdet_upd_opt[k][u] = hparam_sdet_upd[k][u];
      }

      expect_c_opt[k] = expect_c[k];
    }
  }
}

void restore_hparam_opt(void)
{
  int k,v,i,u,n;

  hparam_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class_opt);
  hparam_c_upd = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class_opt);

  if (mode_nominal) {
    hparam_w = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class_opt);
    hparam_w_upd = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class_opt);
    
    for (k=0; k<num_class_opt; k++) {
      hparam_w[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      hparam_w_upd[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
	hparam_w[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
	hparam_w_upd[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
      }
    }
  }

  if (mode_gauss) {
    hparam_u_upd = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class_opt);
    hparam_s_upd = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class_opt);
    hparam_p_upd = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class_opt);
    hparam_sdet_upd = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class_opt);

    expect_c = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class_opt);

    for (k=0; k<num_class_opt; k++) {
      hparam_u_upd[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      hparam_s_upd[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_upd[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_upd[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);

      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];

	hparam_u_upd[k][u] = alloc_vector(n);
	hparam_s_upd[k][u] = alloc_matrix(n,n);
	hparam_p_upd[k][u] = alloc_matrix(n,n);
      }
    }
  }

  for (k=0; k<num_class_opt; k++) {
    hparam_c[k] = hparam_c_opt[k];
    hparam_c_upd[k] = hparam_c_upd_opt[k];
  }

  if (mode_nominal) {
#pragma omp parallel for private(v,i)
    for (k=0; k<num_class_opt; k++) {
      for (v=0; v<tuple_size_d; v++) {
	for (i=0; i<num_word[v]; i++) {
	  hparam_w[k][v][i] = hparam_w_opt[k][v][i];
	  hparam_w_upd[k][v][i] = hparam_w_upd_opt[k][v][i];
	}
      }
    }
  }

  if (mode_gauss) {
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
	copy_vector(hparam_u_upd[k][u],hparam_u_upd_opt[k][u]);
	copy_matrix(hparam_s_upd[k][u],hparam_s_upd_opt[k][u]);
	copy_matrix(hparam_p_upd[k][u],hparam_p_upd_opt[k][u]);
	hparam_sdet_upd[k][u] = hparam_sdet_upd_opt[k][u];
      }

      expect_c[k] = expect_c_opt[k];
    }
  }
}

void release_hparam_opt(void)
{
  int k,v,u;

  free(hparam_c_opt); hparam_c_opt = NULL;

  if (mode_nominal) {
    for (k=0; k<num_class_opt; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(hparam_w_opt[k][v]);     hparam_w_opt[k][v] = NULL;
	free(hparam_w_upd_opt[k][v]); hparam_w_upd_opt[k][v] = NULL;
      }
      free(hparam_w_opt[k]);     hparam_w_opt[k] = NULL;
      free(hparam_w_upd_opt[k]); hparam_w_upd_opt[k] = NULL;
    }
    free(hparam_w_opt);     hparam_w_opt = NULL;
    free(hparam_w_upd_opt); hparam_w_upd_opt = NULL;
  }

  if (mode_gauss) {
    for (k=0; k<num_class_opt; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(hparam_u_upd_opt[k][u]);
	release_matrix(hparam_s_upd_opt[k][u]);
	release_matrix(hparam_p_upd_opt[k][u]);
      }
      free(hparam_u_upd_opt[k]);    hparam_u_upd_opt[k] = NULL;
      free(hparam_s_upd_opt[k]);    hparam_s_upd_opt[k] = NULL;
      free(hparam_p_upd_opt[k]);    hparam_p_upd_opt[k] = NULL;
      free(hparam_sdet_upd_opt[k]); hparam_sdet_upd_opt[k] = NULL;
    }
    free(hparam_u_upd_opt);    hparam_u_upd_opt = NULL;
    free(hparam_s_upd_opt);    hparam_s_upd_opt = NULL;
    free(hparam_p_upd_opt);    hparam_p_upd_opt = NULL;
    free(hparam_sdet_upd_opt); hparam_sdet_upd_opt = NULL;

    free(expect_c_opt);  expect_c_opt = NULL;
  }
}

void release_pi(void)
{
  int k,v;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(pi_w[k][v]); pi_w[k][v] = NULL;
      }
      free(pi_w[k]); pi_w[k] = NULL;
    }
    free(pi_w); pi_w = NULL;
  }

  free(pi_c); pi_c = NULL;
}

void alloc_free_energy_sm_size(void)
{
  int t;

  free_energy_sm_size = (int *)n_malloc(sizeof(int)*(num_sm + 1));
  for (t = 0; t < (num_sm + 1); t++) {
    free_energy_sm_size[t] = MAX_MAX_ITERATE;
  }
}

void alloc_free_energy_sm(void)
{
  int n,t;

  /* We need one extra entry to record the first full EM: */
  free_energy_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*(num_sm + 1));

  for (t=0; t<(num_sm+1); t++) {  /* be careful for the terminal condition */
    n = free_energy_sm_size[t];
    /* n times for partial EM, and n times for full EM, at maximum */
    free_energy_sm[t] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*(2*n));
  }
}

void alloc_log_itemp_sm_vb(void)
{
  int n,t;

  log_itemp_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*(num_sm+1));

  for (t=0; t<(num_sm+1); t++) {
    n = free_energy_sm_size[t];
    log_itemp_sm[t] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*(2*n));
  }
}

void expand_free_energy_sm_all(int t)
{
  int old_free_energy_size;

  old_free_energy_size = free_energy_sm_size[t];
  free_energy_sm_size[t] *= 2;

  free_energy_sm[t] = (FLOAT_T *)n_realloc(free_energy_sm[t],
					    sizeof(FLOAT_T)*free_energy_sm_size[t]);

  if (mode_daem)
    log_itemp_sm[t] = (FLOAT_T *)n_realloc(log_itemp_sm[t],
                                            sizeof(FLOAT_T)*free_energy_sm_size[t]);

  free_energy_best = (FLOAT_T *)n_realloc(free_energy_best,sizeof(FLOAT_T)*free_energy_size);
  
  if (mode_daem)
    log_itemp_best = (FLOAT_T *)n_realloc(log_itemp_best,sizeof(FLOAT_T)*free_energy_size);
}

void release_free_energy_sm_size(void)
{
  free(free_energy_sm_size);
  free_energy_sm_size = NULL;
}

void release_free_energy_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++) {
    free(free_energy_sm[t]); free_energy_sm[t] = NULL;
  }
  free(free_energy_sm); free_energy_sm = NULL;
}

void alloc_hparam_sm(void)
{
  int k,v,u,n;

  hparam_c_sm = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);
  hparam_c_upd_sm = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

  if (mode_nominal) {
    hparam_w_sm = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);
    hparam_w_upd_sm = (FLOAT_T ***)n_malloc(sizeof(FLOAT_T **)*num_class);

    for (k=0; k<num_class; k++) {
      hparam_w_sm[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);
      hparam_w_upd_sm[k] = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
	hparam_w_sm[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
	hparam_w_upd_sm[k][v] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_word[v]);
      }
    }
  }

  if (mode_gauss) {
    hparam_u_sm = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    hparam_s_sm = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p_sm = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

    hparam_u_upd_sm = (VECTOR **)n_malloc(sizeof(VECTOR *)*num_class);
    hparam_s_upd_sm = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_p_upd_sm = (MATRIX **)n_malloc(sizeof(MATRIX *)*num_class);
    hparam_sdet_upd_sm = (FLOAT_T **)n_malloc(sizeof(FLOAT_T *)*num_class);

    expect_c_sm = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*num_class);

    for (k=0; k<num_class; k++) {
      hparam_u_sm[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      hparam_s_sm[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_sm[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_sm[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);

      hparam_u_upd_sm[k] = (VECTOR *)n_malloc(sizeof(VECTOR)*gauss_size);
      hparam_s_upd_sm[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_p_upd_sm[k] = (MATRIX *)n_malloc(sizeof(MATRIX)*gauss_size);
      hparam_sdet_upd_sm[k] = (FLOAT_T *)n_malloc(sizeof(FLOAT_T)*gauss_size);
      
      for (u=0; u<gauss_size; u++) {
	n = gauss_attr_size[u];

	hparam_u_sm[k][u] = alloc_vector(n);
	hparam_s_sm[k][u] = alloc_matrix(n,n);
	hparam_p_sm[k][u] = alloc_matrix(n,n);

	hparam_u_upd_sm[k][u] = alloc_vector(n);
	hparam_s_upd_sm[k][u] = alloc_matrix(n,n);
	hparam_p_upd_sm[k][u] = alloc_matrix(n,n);
      }
    }
  }
}

void release_hparam_sm(void)
{
  int k,v,u;

  if (mode_nominal) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size_d; v++) {
	free(hparam_w_sm[k][v]); hparam_w_sm[k][v] = NULL;
	free(hparam_w_upd_sm[k][v]); hparam_w_upd_sm[k][v] = NULL;
      }
      free(hparam_w_sm[k]); hparam_w_sm[k] = NULL;
      free(hparam_w_upd_sm[k]); hparam_w_upd_sm[k] = NULL;
    }
    free(hparam_w_sm); hparam_w_sm = NULL;
    free(hparam_w_upd_sm); hparam_w_upd_sm = NULL;
  }
    
  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      for (u=0; u<gauss_size; u++) {
	release_vector(hparam_u_sm[k][u]);
	release_matrix(hparam_s_sm[k][u]);
	release_matrix(hparam_p_sm[k][u]);

	release_vector(hparam_u_upd_sm[k][u]);
	release_matrix(hparam_s_upd_sm[k][u]);
	release_matrix(hparam_p_upd_sm[k][u]);
      }

      free(hparam_u_sm[k]);    hparam_u_sm[k] = NULL;
      free(hparam_s_sm[k]);    hparam_s_sm[k] = NULL;
      free(hparam_p_sm[k]);    hparam_p_sm[k] = NULL;
      free(hparam_sdet_sm[k]); hparam_sdet_sm[k] = NULL;

      free(hparam_u_upd_sm[k]);    hparam_u_upd_sm[k] = NULL;
      free(hparam_s_upd_sm[k]);    hparam_s_upd_sm[k] = NULL;
      free(hparam_p_upd_sm[k]);    hparam_p_upd_sm[k] = NULL;
      free(hparam_sdet_upd_sm[k]); hparam_sdet_upd_sm[k] = NULL;
    }

    free(hparam_u_sm);    hparam_u_sm = NULL;
    free(hparam_s_sm);    hparam_s_sm = NULL;
    free(hparam_p_sm);    hparam_p_sm = NULL;
    free(hparam_sdet_sm); hparam_sdet_sm = NULL;

    free(hparam_u_upd_sm);    hparam_u_upd_sm = NULL;
    free(hparam_s_upd_sm);    hparam_s_upd_sm = NULL;
    free(hparam_p_upd_sm);    hparam_p_upd_sm = NULL;
    free(hparam_sdet_upd_sm); hparam_sdet_upd_sm = NULL;

    free(expect_c_sm);  expect_c_sm = NULL;
  }

  free(hparam_c_sm); hparam_c_sm = NULL;
  free(hparam_c_upd_sm); hparam_c_upd_sm = NULL;
}

void release_log_itemp_sm_vb(void)
{
  release_log_itemp_sm();
}
