/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

#if !defined(LANG) || defined(LANG_EN)

char usage_full[] =
#ifdef _OPENMP
    "\n" TOOL_NAME " version " VERSION " (ML+MAP, built on " __DATE__ ", with OpenMP)\n"
#else
    "\n" TOOL_NAME " version " VERSION " (ML+MAP, built on " __DATE__ ")\n"
#endif
    "[Usage] nbc -f basename -k #classes -x #attributes [option...]\n"
    "  Flags with an argument:\n"
    "    --target   (-f) Prefix of input/output files         [Required]\n"
    "    --suffix   (-s) Suffix of output files               [Default: No suffix]\n"
    "    --n-class  (-k) Number of classes or Kmin:Kmax:Kstep [Required unless -R/-H given]\n"
    "    --n-attr   (-x) Number of attributes                 [Required]\n"
    "    --gauss    (-u) Indices of continuous attributes     [Default: No continous attrs]\n"
    "    --id       (-I) Index of the ID attribute\n"
    "    --ignore   (-y) Indices of the attributes to be ignored\n"
    "    --eval     (-E) Index of the class attribute, used for evaluation\n"
    "    --input    (-i) File extension of the input files\n"
    "                    (csv:comma-separated, txt:tab-separated) [Default: automatically detected]\n"
    "    --output   (-j) File extension of the output files\n"
    "                    (csv:comma-separated, tsv:tab-separated) [Default: same as that of inputs]\n"
    "    --smooth-c (-c) Smoothing constant for class parameters\n"
    "                    (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    --smooth-a (-w) Smoothing constants for attribute parameters\n"
    "                    (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    --tau      (-T) Tau in the Normal-Wishart priors\n"
    "                    (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    --alpha    (-O) Alpha in the Normal-Wishart priors\n"
    "    --smooth-mean (-K) Smoothing constant for means\n"
    "    --smooth-var  (-L) Smoothing constant for variances\n"
    "    --init-c Initialization method for class (random:Random, noisy_u:Noisy-uniform)\n"
    "                                                     [Default: Noisy-uniform]\n"
    "    --init-a Initialization method for attributes\n"
    "             (random:Random, noisy_u:Noisy-uniform: v_freq:Value-frequency)\n"
    "                                                     [Default: Noisy-uniform]\n"
    "    --epsilon  (-e) Threshold in judging convergence [Default: 1.0e-6]\n"
    "    --max-iter (-m) Max number of EM iterations      [Default: Iterate EM until convergence]\n"
    "    --score    (-g) Model score (bic:BIC, cs:Cheeseman-Stutz)   [Default: BIC]\n"
    "                    [NOTE] AUTOCLASS is available only when smoothing is enabled\n"
    "    --seed     (-r) Random seed             [Default: return value of time()]\n"
    "    --restart  (-n) Number of restarts      [Default: No restart]\n"
    "    --max-iter-restart (-l) Max number of preliminary EM iterations in random restart\n"
    "                                            [Default: Iterate EM until convergence]\n"
    "    --rank  (-z) Number of attribute values highly relevant to a class, or threshold for\n"
    "                 the relevance (Ranking is enabled)\n"
    "                                            [Default: Ranking is disabled] \n"
    "    --itemp (-B) Binit:Brate\n"
    "                 where Binit is the initial value of the inverse temperature,\n"
    "                 and Brate is the update rate of the inverse temperature    [Default: 0.1:1.5]\n"
    "    --n-sm     (-o) Max number of split-merge operations (SMEM is enabled)  [Default: SMEM is disabled]\n"
    "    --n-cand   (-a) Max number of SM candidates to be examined              [Default: No limit]\n"
    "    --n-jsplit (-G) Max number of split candidates to be examined for each pair being merged\n"
    "                                                                            [Default: No limit]\n"
    "    --min-c-param      Lower limit of prob. of the class to be split/merged [Default: No limit]\n"
    "    --min-new-c-param  Lower limit of prob. of the splited/merged class     [Default: No limit]\n"
    "    --min-sm-imp       Lower limit of improvement by spilt-merge oparations\n"
    "                               [Default: Threshold of convergence (given by -e option)]\n"
#ifdef USE_ICONV
#ifdef CHARCODE
    "    --code-in  (-p) Input character charcode    [Default: " CHARCODE "]\n"
    "    --code-out (-q) Output character charcode   [Default: " CHARCODE "]\n"
    "                    (run 'iconv -l' for available charcodes)\n"
#else
    "    --code-in  (-p) Input character charcode    [Default: " CHARCODE_INTERNAL "]\n"
    "    --code-out (-q) Output character charcode   [Default: " CHARCODE_INTERNAL "]\n"
    "                    (run 'iconv -l' for available charcodes)\n"
#endif
#endif
    "    --prec-out      Number of significant digits of a floating point number\n"
    "                        in the output files      [Default: 12]\n"
    "    --prec-msg      Number of digits after the decimal point of a floating point number\n"
    "                        in the terminal messages [Default: 6]\n"
    "  Flags with no argument:\n"
    "    --read-param  (-R) Disable EM learning and read paramters from a file\n"
    "    --read-hparam (-H) Read hyperparameters from a file\n"
    "    --autoclass   (-A) Set hyperparameters similarly to AutoClass\n"
    "    --mem-prob    (-D) Compute and output the membership distributions\n"
    "    --dissim           Compute and output the dissimilarities\n"
    "    --log-valued  (-J) Enable log-valued computation in ML/MAP estimation\n"
    "    --uncompress  (-U) Uncompress the cases having the same attribute values\n"
    "    --verbose     (-v) Output the temporarily computed results to log file\n"
    "    --daem        (-b) Enable DAEM\n"
    "    --norm-jmerge      Normalize Jmerge\n"
    "    --occ-order        Print attvals in the appearance order     [Default: lexicographical order]\n"
    "    --hide-attr   (-V) Hide attribute values in the output files\n"
    "    --hide-rank        Hide irrevant attribute values in ranking\n"
    "    --save-msg    (-S) Output the terminal messages to the files [Default: to stderr]\n"
    "    --help        (-h) Print brief descriptions on major flags\n"
    "    --helpall     (-X) Print brief descriptions on all flags\n\n";

char usage[] =
#ifdef _OPENMP
    "\n" TOOL_NAME " version " VERSION " (ML+MAP, built on " __DATE__ ", with OpenMP)\n"
#else
    "\n" TOOL_NAME " version " VERSION " (ML+MAP, built on " __DATE__ ")\n"
#endif
    "[Usage] nbc -f basename -k #classes -x #attributes [option...]\n"
    "  Flags with an argument:\n"
    "    --target   (-f) Prefix of input/output files         [Required]\n"
    "    --suffix   (-s) Suffix of output files               [Default: No suffix]\n"
    "    --n-class  (-k) Number of classes or Kmin:Kmax:Kstep [Required unless -R/-H given]\n"
    "    --n-attr   (-x) Number of attributes                 [Required]\n"
    "    --gauss    (-u) Indices of continuous attributes     [Default: No continous attrs]\n"
    "    --id       (-I) Index of the ID attribute\n"
    "    --ignore   (-y) Indices of the attributes to be ignored\n"
    "    --eval     (-E) Index of the class attribute, used for evaluation\n"
    "    --output   (-j) File extension of the output files\n"
    "                    (csv:comma-separated, tsv:tab-separated) [Default: same as that of inputs]\n"
    "    --smooth-c (-c) Smoothing constant for class parameters\n"
    "                    (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    --smooth-a (-w) Smoothing constants for attribute parameters\n"
    "                    (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    --tau      (-T) Tau in the Normal-Wishart priors\n"
    "                    (auto:AutoClass, other positive float: as-is) [Default: No smoothing]\n"
    "    --epsilon  (-e) Threshold in judging convergence                    [Default: 1.0e-6]\n"
    "    --max-iter (-m) Max number of EM iterations      [Default: Iterate until convergence]\n"
    "    --score    (-g) Model score (bic:BIC, cs:Cheeseman-Stutz)              [Default: BIC]\n"
    "                    [NOTE] AUTOCLASS is available only when smoothing is enabled\n"
    "    --restart  (-n) Number of restarts                  [Default: No restart]\n"
    "    --rank     (-z) Number of attribute values highly relevant to a class, or threshold for\n"
    "                    the relevance (Ranking is enabled)  [Default: Ranking is disabled] \n"
#ifdef USE_ICONV
#ifdef CHARCODE
    "    --code-in  (-p) Input character code   [Default: " CHARCODE "]\n"
    "    --code-out (-q) Output character code  [Default: " CHARCODE "]\n"
    "                    (run 'iconv -l' for available code names)\n"
#else
    "    --code-in  (-p) Input character code   [Default: " CHARCODE_INTERNAL "]\n"
    "    --code-out (-q) Output character code  [Default: " CHARCODE_INTERNAL "]\n"
    "                    (run 'iconv -l' for available code names)\n"
#endif
#endif
    "  Flags with no argument:\n"
    "    --read-param  (-R) Disable EM learning and read paramters from a file\n"
    "    --read-hparam (-H) Read hyperparameters from a file\n"
    "    --autoclass   (-A) Set hyperparameters similarly to AutoClass\n"
    "    --log-valued  (-J) Enable log-valued computation in ML/MAP estimation\n"
    "    --save-msg    (-S) Output the messages to the files [Default: to stderr]\n"
    "    --help        (-h) Print brief descriptions on major flags\n"
    "    --helpall     (-X) Print brief descriptions on all flags\n\n";

void print_log_likelihood(FILE *fp, int k, int num_iterate)
{
  int m;

  print_id(fp,"LOGLIKE-K%d",k);

  if (mode_smooth) {
    if (mode_daem) {
      add_desc("Num. of classes");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      add_desc("Log prior prob.");
      add_desc("Log a posteriori prob.");
      add_desc("Inverse temperature");
      print_desc(fp);

      for (m=0; m<num_iterate; m++)
        print_data(fp,"%d,%d,%.Ng,%.Ng,%.Ng,%.Ng",
		   num_class,m,log_likelihood[m],log_prior[m],lambda[m],log_itemp[m]);
    }
    else {
      add_desc("Num. of classes");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      add_desc("Log prior prob.");
      add_desc("Log a posteriori prob.");
      print_desc(fp);
      for (m=0; m<num_iterate; m++)
        print_data(fp,"%d,%d,%.Ng,%.Ng,%.Ng",
		   num_class,m,log_likelihood[m],log_prior[m],lambda[m]);
    }
  }
  else {
    if (mode_daem) {
      add_desc("Num. of classes");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      add_desc("Inverse temperature");
      print_desc(fp);
      for (m=0; m<num_iterate; m++)
        print_data(fp,"%d,%d,%.Ng,%.Ng",num_class,m,log_likelihood[m],log_itemp[m]);
    }
    else {
      add_desc("Num. of classes");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      print_desc(fp);
      for (m=0; m<num_iterate; m++)
        print_data(fp,"%d,%d,%.Ng",num_class,m,log_likelihood[m]);
    }
  }

  fflush(fp);
}

void print_ic(FILE *fp)
{
  int k,u;

  print_id(fp,"SCORE");

  add_desc("Num. of classes");
  add_desc("BIC");
  add_desc("AUTOCLASS");
  print_desc(fp);

  if (mode_search_num_class && !mode_read_hparam)
    print_desc_plain(fp,
		     "Selected criterion = %s, optimal num. of classes = %d",
		     ic_type_str,num_class_opt);
  else
    print_desc_plain(fp,
		     "Selected criterion = %s, optimal num. of classes = %d",
		     ic_type_str,num_class);

  if (mode_search_num_class && !mode_read_hparam) {
    u = 0;
    for (k = kmin; k <= kmax; k += kstep) {
      if (mode_smooth)
	print_data(fp,"%d,%.Ng,%.Ng",k,bic[u],autoclass[u]);
      else
	print_data(fp,"%d,%.Ng,N/A",k,bic[u]);
      u++;
    }
  }
  else {
    if (mode_smooth)
      print_data(fp,"%d,%.Ng,%.Ng",num_class,bic[0],autoclass[0]);
    else
      print_data(fp,"%d,%.Ng,N/A",num_class,bic[0]);
  }
    
  fflush(fp);
}

void print_seed(FILE *fp)
{
  int k,u,r;

  if (mode_search_num_class && !mode_read_hparam) {
    u = 0;
    for (k = kmin; k <= kmax; k += kstep) {
      print_id(fp,"SEED-K%d",k);

      if (mode_smooth) {
	add_desc("Num. of classes");
	add_desc("Restart(best=%d)",log_r_best[u]);
	add_desc("Random seed");
	add_desc("Num. of updates");
	add_desc("Log a posteriori prob.");
      }
      else {
	add_desc("Num. of classes");
	add_desc("Restart(best=%d)",log_r_best[u]);
	add_desc("Random seed");
	add_desc("Num. of updates");
	add_desc("Log likelihood");
      }
      print_desc(fp);

      for (r = 0; r < num_init; r++) {
	if (mode_set_seed)
	  print_data(fp,"%d,%d,N/A,%d,%.Ng",
		     k,r,log_num_iterate[u][r],log_final_lambda[u][r]);
	else
	  print_data(fp,"%d,%d,%u,%d,%.Ng",
		     k,r,log_seed[u][r],log_num_iterate[u][r],log_final_lambda[u][r]);
      }
      u++;
    }
  }
  else {
    print_id(fp,"SEED");

    if (mode_smooth) {
      add_desc("Num. of classes");
      add_desc("Restart(best=%d)",log_r_best[0]);
      add_desc("Random seed");
      add_desc("Num. of updates");
      add_desc("Log a posteriori prob.");
    }
    else {
      add_desc("Num. of classes");
      add_desc("Restart(best=%d)",log_r_best[0]);
      add_desc("Random seed");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
    }
    print_desc(fp);

    for (r = 0; r < num_init; r++) {
      if (mode_set_seed)
	print_data(fp,"%d,%d,N/A,%d,%.Ng",
		   num_class,r,log_num_iterate[0][r],log_final_lambda[0][r]);
      else
	print_data(fp,"%d,%d,%u,%d,%.Ng",
		   num_class,r,log_seed[0][r],log_num_iterate[0][r],log_final_lambda[0][r]);
    }
  }

  fflush(fp);
}

void print_decrease(FILE *fp, int r, int m,
		    FLOAT_T pre_lambda, FLOAT_T post_lambda, FLOAT_T itemp0)
{
  print_id(fp,"ERR-MSG-DECREASE");

  if (mode_smooth)
    print_desc_plain(fp,
		     "Log posterior prob. decreased!");
  else
    print_desc_plain(fp,
		     "Log-likelihood decreased!");

  print_data(fp,"NumberOfClasses,%d",num_class);
  print_data(fp,"Restart,%d",r);
  print_data(fp,"NumberOfUpdates,%d",m);
  if (mode_smooth) {
    print_data(fp,"LogPostProbBeforeUpdate,%.Ng",pre_lambda);
    print_data(fp,"LogPostProbAfterUpdate,%.Ng",post_lambda);
  }
  else {
    print_data(fp,"LogLikelihoodBeforeUpdate,%.Ng",pre_lambda);
    print_data(fp,"LogLikelihoodAfterUpdate,%.Ng",post_lambda);
  }
  if (mode_daem) {
    print_data(fp,"InverseTemperature,%.Ng",itemp0);
  }
}

void print_time(FILE *fp)
{
  print_id(fp,"CPUTIME");

  add_desc("CPU time (sec)");
  print_desc(fp);

  print_data(fp,"Read,%.Ng",(etime_read - stime_read));
  if (mode_em) print_data(fp,"EM,%.Ng",(etime_em - stime_em));
  if (mode_memp) print_data(fp,"MemberDistrib,%.Ng",(etime_memp - stime_memp));
  if (mode_dist) print_data(fp,"KL-Divergence,%.Ng",(etime_kl_dist - stime_kl_dist));

  fflush(fp);

  print_id(fp,"LAPTIME");

  add_desc("Elapsed time (sec)");
  print_desc(fp);

  print_data(fp,"Read,%.Ng",(etime_lap_read - stime_lap_read));
  if (mode_em) print_data(fp,"EM,%.Ng",(etime_lap_em - stime_lap_em));
  if (mode_memp) print_data(fp,"MemberDistrib,%.Ng",(etime_lap_memp - stime_lap_memp));
  if (mode_dist) print_data(fp,"KL-Divergence,%.Ng",(etime_lap_kl_dist - stime_lap_kl_dist));

  fflush(fp);
}

void print_jmerge(FILE *fp)
{
  int k1,k2;

  print_id(fp,"JMERGE");
  
  add_desc("Class 1");
  add_desc("Class 2");
  add_desc("Jmerge");
  print_desc(fp);

  for (k1=0; k1<num_class; k1++)
    for (k2=0; k2<num_class; k2++)
      print_data(fp,"%d,%d,%.Ng",k1,k2,jmerge[k1][k2]);
}

void print_jsplit(FILE *fp)
{
  int k;

  print_id(fp,"JSPLIT");

  add_desc("Class 1");
  add_desc("Class 2");
  add_desc("Jsplit");
  print_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,jsplit[k]);
}

void print_log_likelihood_sm(FILE *fp, int k)
{
  int t,m;

  if (mode_smooth) {
    print_id(fp,"LOGLIKE-SMEM-K%d",k);
    if (mode_daem) {
      add_desc("Num. of classes");
      add_desc("Num. of split-merge operations");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      add_desc("Log prior prob.");
      add_desc("Log a posteriori prob.");
      add_desc("Inverse temperature");
      print_desc(fp);

      for (t=0; t<(num_sm+1); t++)
        for (m=0; m<num_iterate_sm_full[t]; m++)
          print_data(fp,"%d,%d,%d,%.Ng,%.Ng,%.Ng,%.Ng",
		     k,t,m,
		     log_likelihood_sm[t][m],
		     log_prior_sm[t][m],
		     lambda_sm[t][m],
		     log_itemp_sm[t][m]);
    }
    else {
      add_desc("Num. of classes");
      add_desc("Num. of split-merge operations");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      add_desc("Log prior prob.");
      add_desc("Log a posteriori prob.");
      print_desc(fp);

      for (t=0; t<(num_sm+1); t++)
        for (m=0; m<num_iterate_sm_full[t]; m++)
          print_data(fp,"%d,%d,%d,%.Ng,%.Ng,%.Ng",
		     k,t,m,
		     log_likelihood_sm[t][m],
		     log_prior_sm[t][m],
		     lambda_sm[t][m]);
    }
  }
  else {
    print_id(fp,"LOGLIKE-SMEM-K%d",k);
    if (mode_daem) {
      add_desc("Num. of classes");
      add_desc("Num. of split-merge operations");
      add_desc("Num. of updates, Log likelihood");
      add_desc("Inverse temperature");
      print_desc(fp);

      for (t=0; t<(num_sm+1); t++)
        for (m=0; m<num_iterate_sm_full[t]; m++)
          print_data(fp,"%d,%d,%d,%.Ng,%.Ng",
		     k,t,m,log_likelihood_sm[t][m],log_itemp_sm[t][m]);
    }
    else {
      add_desc("Num. of classes");
      add_desc("Num. of split-merge operations");
      add_desc("Num. of updates");
      add_desc("Log likelihood");
      print_desc(fp);

      for (t=0; t<(num_sm+1); t++)
        for (m=0; m<num_iterate_sm_full[t]; m++)
          print_data(fp,"%d,%d,%d,%.Ng",
		     k,t,m,log_likelihood_sm[t][m]);
    }
  }

}

void print_num_result_sm(FILE *fp, int k)
{
  int t;

  print_id(fp,"SMEM-OP-STAT-K%d",k);

  add_desc("Num. of classes");
  add_desc("Num. of split-merge operations");
  add_desc("Num. of rejected candidates");
  add_desc("Num. of ignored candidates");
  print_desc(fp);

  for (t=0; t<num_sm; t++)
    print_data(fp,"%d,%d,%d,%d",k,t,num_result_rejected[t],num_result_skipped[t]);
}

void print_num_iterate_sm(FILE *fp, int k)
{
  int t;

  print_id(fp,"SMEM-EM-STAT-K%d",k);

  add_desc("Num. of classes");
  add_desc("Num. of split-merge operations");
  add_desc("Num. of updates in partial EM");
  add_desc("Num. of updates in full EM");
  print_desc(fp);

  for (t=0; t<(num_sm+1); t++) {
    if (t==0)
      print_data(fp,"%d,%d,N/A,%d",k,t,num_iterate_sm_full[t]);
    else
      print_data(fp,"%d,%d,%d,%d",
		 k,t,num_iterate_sm_part[t],
		 (num_iterate_sm_full[t]-num_iterate_sm_part[t]));
  }

}

void print_decrease_sm(FILE *fp, int t, int m,
                       FLOAT_T pre_lambda, FLOAT_T post_lambda, FLOAT_T itemp0)
{
  print_id(fp,"ERR-MSG-DECREASE");
  if (mode_smooth)
    print_desc_plain(fp,
		     "Log posterior prob. decreased!");
  else
    print_desc_plain(fp,
		     "Log-likelihood decreased!");

  print_data(fp,"NumberOfClasses,%d",num_class);
  print_data(fp,"SplitMergeOperation,%d",t);
  print_data(fp,"NumberOfUpdates,%d",m);
  if (mode_smooth) {
    print_data(fp,"LogPostProbBeforeUpdate,%.Ng",pre_lambda);
    print_data(fp,"LogPostProbAfterUpdate,%.Ng",post_lambda);
  }
  else {
    print_data(fp,"LogLikelihoodBeforeUpdate,%.Ng",pre_lambda);
    print_data(fp,"LogLikelihoodAfterUpdate,%.Ng",post_lambda);
  }
  if (mode_daem) {
    print_data(fp,"InverseTemperature,%.Ng",itemp0);
  }
}

void print_settings(FILE *fp)
{
  print_id(fp,"SETTING");

  add_desc("Settings");
  print_desc(fp);

  print_data(fp,"Version,%s",VERSION);
  print_data(fp,"Case file (input),%s",fname_freq);
  if (mode_read_hparam) print_data(fp,"Hyperparameter file (input),%s",fname_hparam);
  print_data(fp,"Parameter file (output),%s",fname_param);
  print_data(fp,"Log file (output),%s",fname_log);
  if (mode_memp)    print_data(fp,"Member distrib file (output),%s",fname_memp);
  if (mode_dist)    print_data(fp,"Dissimilarity file (output),%s",fname_dist);
  if (mode_cluster) print_data(fp,"Cluster file (output),%s",fname_cluster);
  if (mode_rank)    print_data(fp,"Ranking file (output),%s",fname_rank);
  if (mode_msg)     print_data(fp,"Message log file (output),%s",fname_msg);
  print_data(fp,"Reading parameters,%d",!(mode_em));
  print_data(fp,"Reading hyperparameters,%d",mode_read_hparam);
  print_data(fp,"Member distributions,%d",mode_memp);
  print_data(fp,"Dissimilarities,%d",mode_dist);
  print_data(fp,"Model selection,%d",mode_search_num_class);
  print_data(fp,"Writing attvals in lexicographical order,%d",mode_lex_order);
  print_data(fp,"Saving messages,%d",mode_msg);
  print_data(fp,"Being Verbose,%d",mode_verbose);
  print_data(fp,"Input file extension,%s", ((format_in == FMT_CSV) ?  FEXT_CSV : FEXT_TSV));
  print_data(fp,"Output file extension,%s",((format_out == FMT_CSV) ? FEXT_CSV : FEXT_TSV));
#ifdef USE_ICONV  
  print_data(fp,"Input charcode,%s",charcode_in);
  print_data(fp,"Output charcode,%s",charcode_out);
#endif
  print_data(fp,"Hiding attvals,%d",mode_hide_attr);
  print_data(fp,"Relevant analysis,%d",mode_rank);
  print_data(fp,"Hiding less relevant cases/attvals,%d",mode_rank_hide);
  if (mode_rank_th)
    print_data(fp,"Threshold for relevant cases/attvals,%.Ne",rank_th);
  else 
    print_data(fp,"Num of relevant cases/attvals,%d",num_rank);
  print_data(fp,"Ranking score,%s",rel_m_str);
  print_data(fp,"Max num Of EM iterations,%d",max_iterate);
  print_data(fp,"Threshold for convergence,%.Ne",epsilon);
  print_data(fp,"Initialization method for class,%s",init_type_str_class);
  print_data(fp,"Initialization method for attributes,%s",init_type_str_word);
  if (mode_search_num_class) print_data(fp,"Model score,%s",ic_type_str);
  if (num_init > 1) {
    print_data(fp,"Number of restarts,%d",num_init);
    print_data(fp,"Max num of EM iterations in restarts,%d",max_init_iterate);
  }
  if (num_sm > 0) {
    print_data(fp,"Num of split-merge operations,%d",num_sm);
    print_data(fp,"Max num of split-merge candidates,%d",max_candidate);
    print_data(fp,"Max num of split candidates,%d",max_candidate);
    print_data(fp,"Normalization of J-merge,%d",mode_norm);
    print_data(fp,"Min class prob. before SM op. (0: No limit),%.Ng",min_param_c);
    print_data(fp,"Min class prob. after SM op. (0: No limit),%.Ng",min_new_param_c);
    print_data(fp,"Min improvement,%.Ng",min_smem_improve);
  }
  print_data(fp,"Random seed,%d",rand_seed);
#ifndef NO_LDOUBLE
  print_data(fp,"Byte size (double),%d",sizeof(double));
  print_data(fp,"Byte size (long-double),%d",sizeof(FLOAT_T));
#endif

  fflush(fp);
}

void print_model(FILE *fp)
{
  int v,v_d,u,n,a,first;

  print_id(fp,"MODEL");

  add_desc("Model specification");
  print_desc(fp);

  if (mode_search_num_class && !mode_read_hparam) {
    print_data(fp,"Num of classes (min),%d",kmin);
    print_data(fp,"Num of classes (max),%d",kmax);
    print_data(fp,"Num of classes (step),%d",kstep);
  }
  else {
    print_data(fp,"Num of classes,%d",num_class);
  }
  print_data(fp,"Num of attrs,%d",tuple_size);
  print_data(fp,"Num of attrs (discrete),%d",tuple_size_d);
  print_data(fp,"Num of attrs (continuous),%d",tuple_size_c);
  
  if (mode_id) print_data(fp,"ID attribute,%d",attr_id);
  if (mode_eval) print_data(fp,"Answer class attribute,%d",attr_eval);
  for (v=0; v<tuple_size; v++)
    if (is_attr_ignore(v)) print_data(fp,"Ignored attribute,%d",v);
  print_data_aux(fp,0,"Discrete attrs,");
  first = 1;
  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;
    print_data_aux(fp,0,((first) ? "%d" : " %d"),v);
    first = 0;
  }
  print_data_aux(fp,1,"");

  for (u=0; u<gauss_size; u++) {
    n = gauss_attr_size[u];
    print_data_aux(fp,0,"Attrs in Gauss%d,",u);
    for (a=0; a<n; a++)
      print_data_aux(fp,(a==(n-1)),((a>0)? " %d" : "%d"),gauss_attr[u][a]);
  }

  for (u=0; u<gauss_size; u++)
    print_data(fp,"Dimension of Gauss%d,%d",u,gauss_attr_size[u]);


  if (!mode_read_hparam) {
    if (mode_smooth_auto_c == MODE_SMOOTH_AUTO)
      print_data(fp,"Pseudo count (class),AutoClass");
    else
      print_data(fp,"Pseudo count (class),%.Ng",smooth_c);

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;
      
      v_d = attr_info[v].local_id;
      if (mode_smooth_auto_w[v_d] == MODE_SMOOTH_AUTO)
	print_data(fp,"Pseudo count (attr %d),AutoClass",v);
      else
	print_data(fp,"Pseudo count (attr %d),%.Ng",v,smooth_w[v_d]);
    }

    if (mode_smooth_auto_tau == MODE_SMOOTH_AUTO)
      print_data(fp,"Tau,AutoClass");
    else
      print_data(fp,"Tau,%.Ng",tau);

    if (alpha_w > 0)
      print_data(fp,"Alpha,%.Ng",alpha_w);
    else
      print_data(fp,"Alpha,Dim + noise");

    if (mode_smooth_u)
      print_data(fp,"Mean0,%.Ng",smooth_u);
    else
      print_data(fp,"Mean0,Sample mean");
      
    if (mode_smooth_s)
      print_data(fp,"Sigma0,%.Ng",smooth_s);
    else
      print_data(fp,"Sigma0,Sample variance");
  }

  fflush(fp);
}

void print_attr_list(FILE *fp)
{
  int v,v_d,i,j;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    print_id(fp,"ATTVAL%d",v);

    add_desc("ID for the attribute value");
    add_desc("Attr. value x%d",v);
    print_desc(fp);

    for (i=0; i<num_word[v_d]; i++) {
      j = print_order[v_d][i];
      print_data(fp,"%d,%s",j,word_list[v_d][j]);
    }
  }

  fflush(fp);
}

void print_attr_freq_list(FILE *fp)
{
  int v,v_d,i,j;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    print_id(fp,"FREQ-CAT%d",v);

    add_desc("Attr. ID");
    add_desc("ID of attr. value x%d",v);
    add_desc("Attr. value x%d",v);
    add_desc("Num. of occurrences N(x%d)",v);
    print_desc(fp);

    for (i=0; i<num_word[v_d]; i++) {
      j = print_order[v_d][i];
      print_data(fp,"%d,%d,%s,%d",v,j,word_list[v_d][j],word_freq_list[v_d][j]);
    }
  }

  fflush(fp);
}

void print_case_freq_list(FILE *fp)
{
  int v,e;

  print_id(fp,"FREQ-CASE");

  add_desc("Case ID");
  if (mode_eval) add_desc("Answer class");
  if (!mode_hide_attr) add_desc("Attribute values x");
  add_desc("Num. of occurrences N(x)");
  print_desc(fp);

  for (e=0; e<num_entry; e++) {
    print_data_aux(fp,0,"%s,",get_case_id(e));
    if (mode_eval) print_data_aux(fp,0,"%s,",ans_class_list[freq_list[e].ans_id]);
    if (!mode_hide_attr)
      for (v=0; v<tuple_size; v++)
	if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	  print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
    print_data_aux(fp,1,"%d",freq_list[e].freq);
  }

  fflush(fp);
}

void print_ans_class_list(FILE *fp)
{
  int k;

  print_id(fp,"ANS-CLASS");

  add_desc("Answer class (for evaluation)");
  add_desc("Answer class name");
  print_desc(fp);

  for (k=0; k<num_ans_class; k++) {
    print_data(fp,"%d,%s",k,ans_class_list[k]);
  }
}

void print_smooth(FILE *fp, int k)
{
  int v,v_d;

  print_id(fp,"AUTO-SMOOTH-CLASS-K%d",k);

  add_desc("Num. of classes");
  add_desc("Class smoothing constant (based on AutoClass)");
  print_desc(fp);

  print_data(fp,"%d,%.Ng",k,smooth_c);

  print_id(fp,"AUTO-SMOOTH-K%d",k);

  add_desc("Num. of classes");
  add_desc("Attr. ID");
  add_desc("Attr. smoothing constant (based on AutoClass)");
  print_desc(fp);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;
    v_d = attr_info[v].local_id;
    print_data(fp,"%d,%d,%.Ng",k,v_d,smooth_w[v_d]);
  }
}

void print_size(FILE *fp)
{
  int k,v,v_d,u,n;
  int total_num_param,num_param_d,num_param_c;

  if (mode_nominal) {
    print_id(fp,"VALUESIZE");
    
    add_desc("Attr. ID");
    add_desc("Num. of attribute values");
    print_desc(fp);
    
    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;
      
      v_d = attr_info[v].local_id;
      print_data(fp,"%d,%d",v_d,num_word[v_d]);
    }
  }

  print_id(fp,"DATASIZE");

  add_desc("Total num. of cases");
  add_desc("Num. of entries (original)");
  if (!mode_uncompress) add_desc("Num. of entries (multiple occurences compressed)");
  print_desc(fp);

  if (mode_uncompress)
    print_data(fp,"%d,%d",total_freq,num_entry_pre);
  else
    print_data(fp,"%d,%d,%d",total_freq,num_entry_pre,num_entry);

  print_id(fp,"PARAMSIZE");

  add_desc("Num. of classes");
  add_desc("Num. of class params");
  add_desc("Num. of params for discrete attrs");
  add_desc("Num. of params for continuous attrs");
  add_desc("Total num. of params");
  print_desc(fp);

  if (mode_search_num_class && !mode_read_hparam) {
    for (k = kmin; k <= kmax; k += kstep) {
      num_param_d = 0;
      for (v=0; v<tuple_size_d; v++)
        num_param_d += k * (num_word[v] - 1);

      num_param_c = 0;
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        num_param_c += k * (n + n * (n - 1) / 2);
      }

      total_num_param = (k - 1) + num_param_d + num_param_c;

      print_data(fp,"%d,%d,%d,%d,%d",
                 k,(k-1),num_param_d,num_param_c,total_num_param);
    }
  }
  else {
    k = num_class;

    num_param_d = 0;
    for (v=0; v<tuple_size_d; v++)
      num_param_d += k * (num_word[v] - 1);
    
    num_param_c = 0;
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];
      num_param_c += k * (n + n * (n - 1) / 2);
    }
    
    total_num_param = (k - 1) + num_param_d + num_param_c;
    
    print_data(fp,"%d,%d,%d,%d,%d",
               k,(k-1),num_param_d,num_param_c,total_num_param);
  }
  
  if (total_num_miss_w > 0) {
    print_id(fp,"NUM-MISSING");
    
    add_desc("Attr. ID");
    add_desc("Num. of missing values");
    add_desc("Ratio of missing values");
    print_desc(fp);
    
    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;
      
      v_d = attr_info[v].local_id;
      print_data(fp,"%d,%d,%.Ng",v,num_miss_w[v_d],(FLOAT_T)num_miss_w[v_d]/total_freq);
    }
  }

  fflush(fp);
}

void print_param(FILE *fp)
{
  int k,v,v_d,i,j,u,a,a0,a1;

  print_id(fp,"PARAM-CLASS");

  add_desc("Class k");
  add_desc("Class parameter P(k)");
  print_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,param_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    print_id(fp,"PARAM-CAT%d",v);

    add_desc("Attr. ID");
    add_desc("Class k");
    add_desc("Attr. value x%d",v);
    add_desc("Attr. parameter P(x%d|k)",v);
    print_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
	print_data(fp,"%d,%d,%s,%.Ng",v,k,word_list[v_d][j],param_w[k][v_d][j]);
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      print_id(fp,"PARAM-GAUSS-MEAN%d",u);

      add_desc("Gauss ID");
      add_desc("Class k");
      add_desc("Attr. ID");
      add_desc("Mean");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	for (a=0; a<gauss_attr_size[u]; a++)
	  print_data(fp,"%d,%d,%d,%.Ng",u,k,gauss_attr[u][a],param_u[k][u][a]);
    
      print_id(fp,"PARAM-GAUSS-COVAR%d",u);

      add_desc("Gauss ID");
      add_desc("Class k");
      add_desc("Attr. ID");
      add_desc("Attr. ID");
      add_desc("(Co)Variance");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	for (a0=0; a0<gauss_attr_size[u]; a0++)
	  for (a1=a0; a1<gauss_attr_size[u]; a1++) {
	    print_data(fp,"%d,%d,%d,%d,%.Ng",u,k,gauss_attr[u][a0],gauss_attr[u][a1],param_s[k][u][a0][a1]);
          }
    }
  }

  fflush(fp);
}

void print_hparam(FILE *fp)
{
  int k,v,v_d,i,j;
  int u,n,a,a0,a1;

  print_id(fp,"HPARAM-CLASS");
  
  add_desc("Class k");
  add_desc("Class pseudo count delta(k)");
  print_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,hparam_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;
    
    print_id(fp,"HPARAM-CAT%d",v);

    add_desc("Attr. ID");
    add_desc("Class k");
    add_desc("Attr. value x%d",v);
    add_desc("Attr. pseudo count delta(x%d|k)",v);
    print_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
	print_data(fp,"%d,%d,%s,%.Ng",v,k,word_list[v_d][j],hparam_w[k][v_d][j]);
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      print_id(fp,"HPARAM-GAUSS-MEAN%d",u);
      
      add_desc("Class k");
      add_desc("Gauss ID");
      add_desc("Attr. ID");
      add_desc("Mean");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	for (a=0; a<n; a++)
	  print_data(fp,"%d,%d,%d,%.Ng",k,u,gauss_attr[u][a],hparam_u[k][u][a]);

      print_id(fp,"HPARAM-GAUSS-COVAR%d",u);

      add_desc("Class k");
      add_desc("Gauss ID");
      add_desc("Attr. ID");
      add_desc("Attr. ID");
      add_desc("(Co)Variance");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	for (a0=0; a0<n; a0++)
	  for (a1=a0; a1<n; a1++)
	    print_data(fp,"%d,%d,%d,%d,%.Ng",
		       k,u,gauss_attr[u][a0],gauss_attr[u][a1],
		       hparam_s[k][u][a0][a1]);
    }

    print_id(fp,"HPARAM-GAUSS-TAU");

    add_desc("Class k");
    add_desc("Tau");
    print_desc(fp);

    for (k=0; k<num_class; k++)
      print_data(fp,"%d,%.Ng",k,hparam_tau[k]);

    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      print_id(fp,"HPARAM-GAUSS-ALPHA%d",u);

      add_desc("Class k");
      add_desc("Gauss ID");
      add_desc("Alpha");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	print_data(fp,"%d,%d,%.Ng",k,u,hparam_alpha_w[k][u]);
    }
  }

  fflush(fp);
}

void print_complete_prob(FILE *fp)
{
  int k,v,e;

  print_id(fp,"COMP");

  add_desc("Class k");
  add_desc("Case ID");
  if (!mode_hide_attr) add_desc("Attribute values x");
  add_desc("Complete data prob. P(k,x)");
  print_desc(fp);

  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      print_data_aux(fp,0,"%d,%s,",k,get_case_id(e));
      if (!mode_hide_attr)
	for (v=0; v<tuple_size; v++)
	  if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	    print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
      print_data_aux(fp,1,"%.Ng",prob_comp[k][e]);
    }
  }
  fflush(fp);
}

void print_incomplete_prob(FILE *fp)
{
  int v,e;

  print_id(fp,"INCOMP");

  add_desc("Case ID");
  if (!mode_hide_attr) add_desc("Attribute values x");
  add_desc("Incomplete data prob. P(x)");
  print_desc(fp);

  for (e=0; e<num_entry; e++) {
    print_data_aux(fp,0,"%s,",get_case_id(e));
    if (!mode_hide_attr)
      for (v=0; v<tuple_size; v++)
	if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	  print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
    print_data_aux(fp,1,"%.Ng",prob_incomp[e]);
  }

  fflush(fp);
}

void print_conditional_prob(FILE *fp)
{
  int k,e,v;

  print_id(fp,"COND");

  add_desc("Case ID");
  add_desc("Attribute values x");
  add_desc("Class k, Conditional prob. P(k|x)");
  print_desc(fp);

  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      print_data_aux(fp,0,"%s,",get_case_id(e));
      for (v=0; v<tuple_size; v++)
	if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	  print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
      print_data_aux(fp,1,"%d,%.Ng",k,prob_cond[k][e]);
    }
  }

  fflush(fp);
}

void print_expectation(FILE *fp)
{
  int k,v,v_d,i,j,u,a,a0,a1;

  print_id(fp,"EXPECT-CLASS");

  add_desc("Class k");
  add_desc("Expected occurrences E[k]");
  print_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,expect_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    print_id(fp,"EXPECT-CAT%d",v);

    add_desc("Attr ID.");
    add_desc("Class k");
    add_desc("Attr. value x%d",v);
    add_desc("Expected occurrences E[x%d|k]",v);
    print_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
	print_data(fp,"%d,%d,%s,%.Ng",v,k,word_list[v_d][j],expect_w[k][v_d][j]);
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      print_id(fp,"EXPECT-GAUSS-MEAN%d",u);

      add_desc("Class k");
      add_desc("Gauss ID");
      add_desc("Attr. ID");
      add_desc("Mean");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	for (a=0; a<gauss_attr_size[u]; a++)
	  print_data(fp,"%d,%d,%d,%.Ng",k,u,gauss_attr[u][a],expect_u[k][u][a]);
      
      print_id(fp,"EXPECT-GAUSS-COVAR%d",u);

      add_desc("Class k");
      add_desc("Gauss ID");
      add_desc("Attr. ID");
      add_desc("Attr. ID");
      add_desc("(Co)Variance");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	for (a0=0; a0<gauss_attr_size[u]; a0++)
	  for (a1=a0; a1<gauss_attr_size[u]; a1++)
	    print_data(fp,"%d,%d,%d,%d,%.Ng",
		       k,u,gauss_attr[u][a0],gauss_attr[u][a1],expect_s[k][u][a0][a1]);
    }
  }

  fflush(fp);
}

void print_case_member_prob(FILE *fp)
{
  int k,v,e;

  print_id(fp,"MEMBER-CASE");

  add_desc("Case ID");
  if (mode_eval) add_desc("Answer class");
  if (!mode_hide_attr) add_desc("Attribute values x");
  add_desc("Class k");
  add_desc("Membership prob. P(k|x)");
  print_desc(fp);

  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      print_data_aux(fp,0,"%s,",get_case_id(e));
      if (mode_eval)
	print_data_aux(fp,0,"%s,",ans_class_list[freq_list[e].ans_id]);
      if (!mode_hide_attr)
	for (v=0; v<tuple_size; v++)
	  if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	    print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
      print_data_aux(fp,1,"%d,%.Ng",k,prob_mem_case[k][e]);
    }
  }

  fflush(fp);
}

void print_attr_member_prob(FILE *fp)
{
  int k,v,i,v_d,j,u,e,a;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    print_id(fp,"MEMBER-CAT%d",v);

    add_desc("Attr. ID");
    add_desc("Attr. value x%d",v);
    add_desc("Class k");
    add_desc("Membership prob. P(k|x%d)",v);
    print_desc(fp);

    for (i=0; i<num_word[v_d]; i++) {
      j = print_order[v_d][i];
      for (k=0; k<num_class; k++)
	print_data(fp,"%d,%s,%d,%.Ng",v,word_list[v_d][j],k,prob_mem_w[k][v_d][j]);
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      print_id(fp,"MEMBER-GAUSS%d",u);
      
      add_desc("Gauss ID");
      add_desc("Case ID");
      if (!mode_hide_attr)
	for (a=0; a<gauss_attr_size[u]; a++)
	  add_desc("Attr. x%d",gauss_attr[u][a]);
      add_desc("Class k");
      add_desc("Membership prob. P(k|Gauss%d)",u);
      print_desc(fp);
      
      for (e=0; e<num_entry; e++) {
	for (k=0; k<num_class; k++) {
	  print_data_aux(fp,0,"%d,%s,",u,get_case_id(e));
	  if (!mode_hide_attr)
	    for (a=0; a<gauss_attr_size[u]; a++)
	      print_data_aux(fp,0,"%s,",freq_entry_val(e,gauss_attr[u][a]));
	  print_data_aux(fp,1,"%d,%.Ng",k,prob_mem_g[k][u][e]);
	}
      }
    }
  }

  fflush(fp);
}

void print_cluster(FILE *fp)
{
  int k,t,i,j,v,v_d,e,u,a;

  switch (cl_type) {

  case CL_MP:
    print_id(fp,"CLUSTER-CASE-BY-CASE");

    add_desc("Case ID");
    if (mode_eval) add_desc("Answer class");
    add_desc("Cluster k");
    if (!mode_hide_attr)
      add_desc("Attribute values x");
    add_desc("Membership prob. P(k|x)");
    print_desc(fp);

    for (e=0; e<num_entry; e++) {
      k = cluster_case_pred[e];

      print_data_aux(fp,0,"%s,",get_case_id(e));
      if (mode_eval)
	print_data_aux(fp,0,"%s,",ans_class_list[freq_list[e].ans_id]);
      print_data_aux(fp,0,"%d,",k);
      if (!mode_hide_attr)
	for (v=0; v<tuple_size; v++)
	  if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	    print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
      print_data_aux(fp,1,"%.Ng",prob_mem_case[k][e]);
    }

    print_id(fp,"CLUSTER-CASE-BY-CLUSTER");

    add_desc("Cluster k");
    if (mode_eval) add_desc("Answer class");
    add_desc("Case ID");
    if (!mode_hide_attr) add_desc("Attribute values x");
    add_desc("Membership prob. P(k|x)");
    print_desc(fp);

    for (k=0; k<num_class; k++) {
      for (t=0; t<cluster_case_freq[k]; t++) {
        e = cluster_case[k][t];

	print_data_aux(fp,0,"%d,",k);
	if (mode_eval) print_data_aux(fp,0,"%s,",ans_class_list[freq_list[e].ans_id]);
	print_data_aux(fp,0,"%s,",get_case_id(e));
	if (!mode_hide_attr)
	  for (v=0; v<tuple_size; v++)
	    if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	      print_data_aux(fp,0,"%s,",freq_entry_val(e,v));
        print_data_aux(fp,1,"%.Ng",prob_mem_case[k][e]);
      }
    }
    break;

  default:
    die("Internal error: Unsupported clustering method\n");
  }

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    switch (cl_type) {

    case CL_MP:
      print_id(fp,"CLUSTER-CAT%d-BY-ATT",v);

      add_desc("Attr. ID");
      add_desc("Attr. x%d",v);
      add_desc("Cluster k");
      add_desc("Membership prob. P(k|x%d)",v);
      print_desc(fp);

      for (i=0; i<num_word[v_d]; i++) {
	j = print_order[v_d][i];
	k = cluster_w_pred[v_d][j];
	print_data(fp,"%d,%s,%d,%.Ng",v,word_list[v_d][j],k,prob_mem_w[k][v_d][j]);
      }

      print_id(fp,"CLUSTER-CAT%d-BY-CLUSTER",v);

      add_desc("Attr. ID");
      add_desc("Cluster k");
      add_desc("Attr. x%d",v);
      add_desc("Membership prob. P(k|x%d)",v);
      print_desc(fp);

      for (k=0; k<num_class; k++) {
	for (t=0; t<cluster_w_freq[v_d][k]; t++) {
	  i = cluster_w[v_d][k][t];
	  print_data(fp,"%d,%d,%s,%.Ng",
		     v,k,word_list[v_d][i],prob_mem_w[k][v_d][i]);
	}
      }
      break;

    default:
      die("Internal error: Unsupported clustering method\n");
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      switch (cl_type) {
	
      case CL_MP:
	print_id(fp,"CLUSTER-GAUSS%d-BY-CASE",u);
	
	add_desc("Gauss ID");
	add_desc("Case ID");
	if (!mode_hide_attr)
	  for (a=0; a<gauss_attr_size[u]; a++)
	    add_desc("Attr. x%d",gauss_attr[u][a]);
	add_desc("Cluster k");
	add_desc("Membership prob. P(k|Gauss%d)",u);
	print_desc(fp);
	
	for (e=0; e<num_entry; e++) {
	  k = cluster_g_pred[u][e];
	  print_data_aux(fp,0,"%d,%s,",u,get_case_id(e));
	  if (!mode_hide_attr)
	    for (a=0; a<gauss_attr_size[u]; a++)
	      print_data_aux(fp,0,"%s,",freq_entry_val(e,gauss_attr[u][a]));
	  print_data_aux(fp,1,"%d,%.Ng",k,prob_mem_g[k][u][e]);
	}
	
	print_id(fp,"CLUSTER-GAUSS%d-BY-CLUSTER",u);
	
	add_desc("Gauss ID");
	add_desc("Cluster k");
	add_desc("Case ID");
	if (!mode_hide_attr)
	  for (a=0; a<gauss_attr_size[u]; a++)
	    add_desc("Attr. x%d",gauss_attr[u][a]);
	add_desc("Membership prob. P(k|Gauss%d)",u);
	print_desc(fp);
	
	for (k=0; k<num_class; k++) {
	  for (t=0; t<cluster_g_freq[u][k]; t++) {
	    e = cluster_g[u][k][t];
	    print_data_aux(fp,0,"%d,%d,%s,",u,k,get_case_id(e));
	    if (!mode_hide_attr)
	      for (a=0; a<gauss_attr_size[u]; a++)
		print_data_aux(fp,0,"%s,",freq_entry_val(e,gauss_attr[u][a]));
	    print_data_aux(fp,1,"%.Ng",prob_mem_g[k][u][e]);
	  }
	}
	break;
	
      default:
	die("Internal error: Unsupported clustering method\n");
      }
    }
  }

  fflush(fp);
}


void print_cluster_freq(FILE *fp)
{
  int k,v,v_d,u;

  print_id(fp,"CLUSTER-FREQ-CASE");

  add_desc("Cluster for cases");
  add_desc("Cluster size");
  print_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%d",k,cluster_case_freq[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    print_id(fp,"CLUSTER-FREQ-CAT%d",v);

    add_desc("Attr. ID");
    add_desc("Cluster for attr. %d",v);
    add_desc("Cluster size");
    print_desc(fp);

    for (k=0; k<num_class; k++)
      print_data(fp,"%d,%d,%d",v,k,cluster_w_freq[v_d][k]);
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      print_id(fp,"CLUSTER-FREQ-GAUSS%d",u);

      add_desc("Gauss ID");
      add_desc("Cluster for Gauss%d",u);
      add_desc("Cluster size");
      print_desc(fp);

      for (k=0; k<num_class; k++)
	print_data(fp,"%d,%d,%d",u,k,cluster_g_freq[u][k]);
    }
  }

  fflush(fp);
}

void print_rank_ml(FILE *fp)
{
  int k,k0,i,j,v,v_d,e,u,a;
  int num_show_rank_case;
  int *num_show_rank_w;

  num_show_rank_case =
    (!mode_rank_th && num_rank > 0 && num_rank < num_entry) ? num_rank : num_entry;

  print_id(fp,"RANK-CASE");
 
  add_desc("Class k");
  if (mode_eval) add_desc("Answer class");
  add_desc("Case ID");
  if (!mode_hide_attr) add_desc("Attributes values x");
  add_desc("Class prob. P(k)");
  add_desc("Membership prob. P(k|x)");
  print_desc(fp);
  print_desc_plain(fp,"(in the decending order of the class probs. and the membership probs.)");;

  for (k=0; k<num_class; k++) {
    k0 = rank_class[k];
    for (e=0; e<num_show_rank_case; e++) {
      j = rank_mem_case[k0][e];
      if (mode_rank_th && prob_mem_case[k0][j] < rank_th) break;
      if (mode_rank_hide && prob_mem_case[k0][j] < 1.0/num_class) break;

      print_data_aux(fp,0,"%d,",k0);
      if (mode_eval) print_data_aux(fp,0,"%s,",ans_class_list[freq_list[j].ans_id]);
      print_data_aux(fp,0,"%s,",get_case_id(j));
      if (!mode_hide_attr)
	for (v=0; v<tuple_size; v++)
	  if (!is_attr_id(v) && !is_attr_eval(v) && !is_attr_ignore(v))
	    print_data_aux(fp,0,"%s,",freq_entry_val(j,v));
      print_data_aux(fp,1,"%.Ng,%.Ng",param_c[k0],prob_mem_case[k0][j]);
    }
  }

  if (mode_nominal) {
    num_show_rank_w = (int *)n_malloc(sizeof(int)*tuple_size_d);
    
    for (v=0; v<tuple_size_d; v++)
      num_show_rank_w[v] =
	(!mode_rank_th && num_rank > 0 && num_rank < num_word[v]) ? num_rank : num_word[v];

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;
      
      v_d = attr_info[v].local_id;
      print_id(fp,"RANK-CAT%d",v);

      add_desc("Attr. ID");
      add_desc("Class k");
      add_desc("Attr. value x%d",v);
      add_desc("Class prob. P(k)");
      add_desc("Membership prob. P(k|x%d)",v);
      print_desc(fp);
      print_desc_plain(fp,
		       "(in the decending order of the class probs. and the membership probs.)");

      for (k=0; k<num_class; k++) {
	k0 = rank_class[k];
	for (i=0; i<num_show_rank_w[v_d]; i++) {
	  j = rank_mem_w[k0][v_d][i];
	  if (mode_rank_th && prob_mem_w[k0][v_d][j] < rank_th) break;
	  if (mode_rank_hide && prob_mem_w[k0][v_d][j] < 1.0/num_class) break;

	  print_data(fp,"%d,%d,%s,%.Ng,%.Ng",
		     v,k0,word_list[v_d][j],param_c[k0],prob_mem_w[k0][v_d][j]);
	}
      }
    }
    
    free(num_show_rank_w);
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      print_id(fp,"RANK-GAUSS%d",u);
    
      add_desc("Gauss ID");
      add_desc("Class k");
      add_desc("Case ID");
      if (!mode_hide_attr)
	for (a=0; a<gauss_attr_size[u]; a++)
	  add_desc("Attr. x%d",gauss_attr[u][a]);
      add_desc("Class prob. P(k)");
      add_desc("Membership prob. P(k|Gauss%d)",u);
      print_desc(fp);
      print_desc_plain(fp,
		       "(in the decending order of the class probs. and the membership probs.)");

      for (k=0; k<num_class; k++) {
	k0 = rank_class[k];
	for (e=0; e<num_show_rank_case; e++) {
	  j = rank_mem_g[k0][u][e];
	  if (mode_rank_th && prob_mem_g[k0][u][j] < rank_th) break;
	  if (mode_rank_hide && prob_mem_g[k0][u][j] < 1.0/num_class) break;

	  print_data_aux(fp,0,"%d,%d,%s,",u,k0,get_case_id(j));
	  if (!mode_hide_attr)
	    for (a=0; a<gauss_attr_size[u]; a++)
	      print_data_aux(fp,0,"%s,",freq_entry_val(j,gauss_attr[u][a]));
	  print_data_aux(fp,1,"%.Ng,%.Ng",param_c[k0],prob_mem_g[k0][u][j]);
	}
      }
    }
  }

  fflush(fp);
}

void print_rank(FILE *fp)
{
  switch (rel_m) {
  case REL_MP:
    print_rank_ml(fp);
    break;
  default:
    fprintf(stderr,"Internal error: Unexpected relevance measure between class and attribute values\n");
    exit(1);
  }
}

void print_dissim_case_header(FILE *fp)
{
  if (!mode_eval) {
    print_id(fp,"DISSIM-CASE");

    add_desc("Case ID of x");
    add_desc("Case ID of x'");
    add_desc("KL-divergence D(x||x')");
    print_desc(fp);
  }
  else {
    print_id(fp,"DISSIM-CASE");

    add_desc("Case ID of x");
    add_desc("Answer class of x");
    add_desc("Case ID of x'");
    add_desc("Answer class of x'");
    add_desc("KL-divergence D(x||x')");
    print_desc(fp);
  }
}

void print_dissim_w_header(FILE *fp, int v)
{
  print_id(fp,"DISSIM-CAT%d",v);

  add_desc("Attr. ID");
  add_desc("Attr. value x%d",v);
  add_desc("Attr. value x%d'",v);
  add_desc("KL-divergence D(x%d||x%d')",v,v);
  print_desc(fp);
}

void print_dissim_g_header(FILE *fp, int u)
{
  int n = gauss_attr_size[u], a;

  print_id(fp,"DISSIM-GAUSS%d",u);

  add_desc("Gauss ID");

  add_desc("Case ID of Gauss%d",u);
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      add_desc("Attr. x%d",gauss_attr[u][a]);

  add_desc("Case ID of Gauss%d'",u);
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      add_desc("Attr. x%d'",gauss_attr[u][a]);
  add_desc("KL-divergence D(Gauss%d||Gauss%d')",u,u);

  print_desc(fp);
}

void print_dissim_case(FILE *fp, int e0, int e1, FLOAT_T d)
{
  if (!mode_eval)
    print_data(fp,"%s,%s,%.Ng",get_case_id(e0),get_case_id(e1),d);
  else
    print_data(fp,"%s,%s,%s,%s,%.Ng",
	       get_case_id(e0),ans_class_list[freq_list[e0].ans_id],
	       get_case_id(e1),ans_class_list[freq_list[e1].ans_id],
	       d);
}

void print_dissim_w(FILE *fp, int v, int v_d, int i, int j, FLOAT_T d)
{
  print_data(fp,"%d,%s,%s,%.Ng",v,word_list[v_d][i],word_list[v_d][j],d);
}

void print_dissim_g(FILE *fp, int u, int e0, int e1, FLOAT_T d)
{
  int n = gauss_attr_size[u], a;

  print_data_aux(fp,0,"%d,",u);
  
  print_data_aux(fp,0,"%s,",get_case_id(e0));
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      print_data_aux(fp,0,"%s,",freq_entry_val(e0,gauss_attr[u][a]));

  print_data_aux(fp,0,"%s,",get_case_id(e1));
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      print_data_aux(fp,0,"%s,",freq_entry_val(e1,gauss_attr[u][a]));

  print_data_aux(fp,1,"%.Ng",d);
}

#endif
