/*
 * NBCT
 * Copyright (c) 2007-2008, Yoshitaka Kameya
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_var.h"
#include "vnbc_def.h"
#include "vnbc_var.h"

static int num_smooth_w = 0;

static void initialize_mode()
{
  fname_base[0] = '\0';

  num_class = 0;
  tuple_size_d = 0;
  kmin = -1;
  kmax = -1;
  kstep = -1;
  epsilon = DEFAULT_EPSILON;
  max_iterate = DEFAULT_MAX_ITERATE;
  rand_seed = DEFAULT_RAND_SEED;
  smooth_c = 0.0;
  smooth_w = NULL;
  ic_type = VB;
  rel_m = REL_MP;
  init_type_class = NOISY_UNIFORM_C;
  init_type_word = NOISY_UNIFORM_W;
  mode_verbose = 0;
  mode_msg = 0;
  format_in = -1;
  format_out = -1;
#ifdef USE_ICONV
#ifdef CHARCODE
  strncpy(charcode_in,CHARCODE,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE,CHARCODE_LEN);
#else
  strncpy(charcode_in,CHARCODE_INTERNAL,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE_INTERNAL,CHARCODE_LEN);
#endif
#endif
  mode_em = 1;
  mode_smooth = 0;
  mode_smooth_strict = 0;
  mode_smooth_auto = 0;
  mode_smooth_auto_c = MODE_SMOOTH_NONE;
  mode_smooth_auto_w = NULL;
  mode_smooth_auto_tau = MODE_SMOOTH_NONE;
  mode_bdeu = 0;
  mode_memp = 0;
  mode_cluster = 1;
  mode_dist = 0;
  mode_as_is = 0;
  mode_eval = 0;
  attr_eval = -1;
  mode_search_num_class = 0;
  mode_suffix = 0;
  num_init = 1;
  max_init_iterate = 0;
  mode_set_seed = 0;
  mode_lex_order = 1;
  mode_rank = 0;
  mode_rank_hide = 0;
  mode_rank_th = 0;
  num_rank = -1;
  rank_th = -1;
  num_sm = 0;
  max_candidate = 0;
  max_jsplit = 0;
  mode_norm = 0;
  min_param_c = DEFAULT_MIN_PARAM_C;
  min_smem_improve = -1;
  min_new_param_c = DEFAULT_MIN_NEW_PARAM_C;
  mode_daem = 0;
  smooth_w = NULL;
  itemp_init = DEFAULT_ITEMP_INIT;
  itemp_rate = DEFAULT_ITEMP_RATE;
  mode_gauss = 0;
  gauss_list = NULL;
  mode_smooth_u = 0;
  smooth_u = 0;
  smooth_s = 0;
  tau = 0;
  alpha_w = 0;
  mode_id = 0;
  attr_id = -1;
  mode_ignore = 0;
  attr_ignore = NULL;
  mode_hide_attr = 0;
  cl_type = CL_MP;
  prec_out = DEFAULT_PREC_OUT;
  prec_msg = DEFAULT_PREC_MSG;
  mode_uncompress = 0;

#ifdef USE_ICONV
#  ifdef CHARCODE
  mode_sys_out  = (strcmp(CHARCODE,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#  endif
#endif
}

static void build_filename(void)
{
  char tmp_fname_tsv[MAX_FILENAME_AUX];
  char tmp_fname_csv[MAX_FILENAME_AUX];

  strcpy(fname_freq,fname_base);
  strcpy(fname_param,fname_base);
  strcpy(fname_hparam,fname_base);
  strcpy(fname_dist,fname_base);
  strcpy(fname_memp,fname_base);
  strcpy(fname_log,fname_base);
  strcpy(fname_msg,fname_base);
  strcpy(fname_cluster,fname_base);
  strcpy(fname_rank,fname_base);
  strcpy(fname_rank_sm,fname_base);
  if (mode_suffix) {
    strcat(fname_param,"_");   strcat(fname_param, fname_suffix);
    strcat(fname_hparam,"_");  strcat(fname_hparam,fname_suffix);
    strcat(fname_dist, "_");   strcat(fname_dist,  fname_suffix);
    strcat(fname_memp, "_");   strcat(fname_memp,  fname_suffix);
    strcat(fname_log,  "_");   strcat(fname_log,   fname_suffix);
    strcat(fname_msg,  "_");   strcat(fname_msg,   fname_suffix);
    strcat(fname_cluster,"_"); strcat(fname_cluster, fname_suffix);
    strcat(fname_rank,"_");    strcat(fname_rank, fname_suffix);
    strcat(fname_rank_sm,"_"); strcat(fname_rank_sm, fname_suffix);
  }

  /* Build the filename of the input */
  if (mode_em) {
    strcpy(tmp_fname_tsv, fname_freq);
    strcpy(tmp_fname_csv, fname_freq);
    strcat(tmp_fname_tsv, "_" FSX_FREQ "." FEXT_TSV);
    strcat(tmp_fname_csv, "_" FSX_FREQ "." FEXT_CSV);
  }
  else {
    strcpy(tmp_fname_tsv,fname_hparam);
    strcpy(tmp_fname_csv,fname_hparam);
    strcat(tmp_fname_tsv, "_" FSX_HPARAM "." FEXT_TSV);
    strcat(tmp_fname_csv, "_" FSX_HPARAM "." FEXT_CSV);
  }

  /* format_in and format_out will be fixed here */
  determine_format(tmp_fname_tsv,tmp_fname_csv);

  strcat(fname_freq,  ((format_in == FMT_CSV) ?  "_" FSX_FREQ  "." FEXT_CSV  : "_" FSX_FREQ  "." FEXT_TSV));
  strcat(fname_param, ((format_out == FMT_CSV) ? "_" FSX_PARAM "." FEXT_CSV  : "_" FSX_PARAM "." FEXT_TSV));
  strcat(fname_hparam, (((mode_em && format_out == FMT_CSV) || (!mode_em && format_in == FMT_CSV)) ?
		       "_" FSX_HPARAM "." FEXT_CSV : "_" FSX_HPARAM "." FEXT_TSV));
  strcat(fname_dist, ((format_out == FMT_CSV) ? "_" FSX_DIST_VB "." FEXT_CSV  : "_" FSX_DIST_VB "." FEXT_TSV));
  strcat(fname_memp, ((format_out == FMT_CSV) ? "_" FSX_MEMP_VB "." FEXT_CSV  : "_" FSX_MEMP_VB "." FEXT_TSV));
  strcat(fname_log,  ((format_out == FMT_CSV) ? "_" FSX_LOG_VB  "." FEXT_CSV  : "_" FSX_LOG_VB  "." FEXT_TSV));
  strcat(fname_msg,  "_" FSX_MSG_VB "." FEXT_TSV);
  strcat(fname_cluster, ((format_out == FMT_CSV) ? "_" FSX_CLUSTER_VB "." FEXT_CSV : "_" FSX_CLUSTER_VB "." FEXT_TSV));
  strcat(fname_rank,    ((format_out == FMT_CSV) ? "_" FSX_RANK_VB    "." FEXT_CSV : "_" FSX_RANK_VB    "." FEXT_TSV));
  strcat(fname_rank_sm, ((format_out == FMT_CSV) ? "_" FSX_SMEM_VB    "." FEXT_CSV : "_" FSX_SMEM_VB    "." FEXT_TSV));
}

static void parse_args(int argc, char *argv[])
{
  int c;

  /* Firstly check -h and the options that relate to log files (filenames etc.)*/
  init_getopt();
  while (1) {
    c = getopt(argc,argv,vnbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case GETOPT_ERR:
      fprintf(stderr,"Unknown/illegal option specification: %s (Use -h for usage)\n",optstr);
      exit(1);
    case O_VB_TARGET:
      strncpy(fname_base,optarg,MAX_FILENAME);
      break;
    case O_VB_HELP:
      print_usage(stdout);
      exit(0);
    case O_VB_XHELP:
      print_usage_full(stdout);
      exit(0);
    default:
      break;
    }
  }

  if (fname_base[0] == '\0') {
    fprintf(stderr,"Missing or illegal --target (-f) option (Use -h for usage)\n");
    exit(1);
  }

  init_getopt();
  while (1) {
    c = getopt(argc,argv,vnbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case GETOPT_ERR:
      fprintf(stderr,"Unknown/illegal option specification: %s (Use -h for usage)\n",optstr);
      exit(1);
    case O_VB_INPUT:
      tolower_str(optarg_c);
      if (strcmp(optarg_c,FEXT_CSV) == 0) {
	format_in = FMT_CSV;
      }
      else if (strcmp(optarg_c,FEXT_TSV) == 0) {
	format_in = FMT_TSV;
      }
      else {
	fprintf(stderr,"Unexpected file format: %s (Use -h for usage)\n",optarg);
	exit(1);
      }
      break;
    case O_VB_OUTPUT:
      tolower_str(optarg_c);
      if (strcmp(optarg_c,FEXT_CSV) == 0) {
	format_out = FMT_CSV;
      }
      else if (strcmp(optarg_c,FEXT_TSV) == 0) {
	format_out = FMT_TSV;
      }
      else {
	fprintf(stderr,"Unexpected file format: %s (Use -h for usage)\n",optarg);
	exit(1);
      }
      break;
    case O_VB_SUFFIX:
      mode_suffix = 1;
      strncpy(fname_suffix,optarg,MAX_FILENAME);
      break;
    case O_VB_SAVE_MESSAGE:
      mode_msg = 1;
      break;
    case O_VB_READ_HPARAM:
      mode_em = 0;
      break;
    default:
      break;
    }
  }

  build_filename();

  /* hereafter we can use the die() family and n_malloc() */
  open_log();

  /* Check --n-attr (-x) option, which is referred to by --ignore (-y) option */
  init_getopt();
  while (1) {
    c = getopt(argc,argv,nbc_opts);
    if (c == GETOPT_FIN) break;
    
    switch (c) {
    case O_N_ATTR:
      tuple_size = get_int(optarg);
      break;
    default:
      break;
    }
  }
  if (tuple_size < 1)
    die("Missing or illegal --n-attr (-x) option (Use -h for usage)\n");

  init_getopt();
  while (1) {
    c = getopt(argc,argv,vnbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case O_VB_EVAL:
      mode_eval = 1;
      attr_eval = get_int(optarg);
      break;
    case O_VB_ID:
      mode_id = 1;
      attr_id = get_int(optarg);
      break;
    default:
      break;
    }
  }

  init_getopt();
  while (1) {
    c = getopt(argc,argv,vnbc_opts);
    if (c == GETOPT_FIN) break;

    switch (c) {
    case O_VB_N_CAND:
      max_candidate = get_int(optarg);
      break;
    case O_VB_DAEM:
      mode_daem = 1;
      break;
    case O_VB_ITEMP:
      get_itemp(optarg);
      break;
    case O_VB_SMOOTH_C:
      smooth_c = get_smooth_c(optarg);
      break;
    case O_VB_MIN_NEW_C_PARAM:
      min_new_param_c = get_float(optarg);
      break;
    case O_VB_DISSIM:
      mode_dist = 1;
      break;
    case O_VB_MEM_PROB:
      mode_memp = 1;
      break;
    case O_VB_EPSILON:
      epsilon = get_float(optarg);
      break;
    case O_VB_NORM_JMERGE:
      mode_norm = 1;
      break;
    case O_VB_N_JSPLIT:
      max_jsplit = get_int(optarg);
      break;
    case O_VB_N_CLASS:
      get_num_class(optarg);  /* we need a special care */
      break;
    case O_VB_SMOOTH_MEAN:
      smooth_u = get_float(optarg);
      mode_smooth_u = 1;
      break;
    case O_VB_MAX_ITER_RESTART:
      max_init_iterate = get_int(optarg);
      break;
    case O_VB_SMOOTH_VAR:
      smooth_s = get_float(optarg);
      mode_smooth_s = 1;
      break;
    case O_VB_MAX_ITER:
      max_iterate = get_int(optarg);
      break;
    case O_VB_RESTART:
      num_init = get_int(optarg);
      break;
    case O_VB_MIN_SM_IMP:
      min_smem_improve = get_float(optarg);
      break;
    case O_VB_N_SM:
      num_sm = get_int(optarg);
      break;
    case O_VB_ALPHA:
      alpha_w = get_float(optarg);
      break;
#ifdef USE_ICONV
    case O_VB_CODE_IN:
      strncpy(charcode_in,optarg,CHARCODE_LEN);
      break;
    case O_VB_CODE_OUT:
      strncpy(charcode_out,optarg,CHARCODE_LEN);
      break;
#endif
    case O_VB_SEED:
      rand_seed = get_uint(optarg);
      mode_set_seed = 1;
      break;
    case O_VB_GAUSS:
      mode_gauss = 1;
      add_gauss(optarg);
      break;
    case O_VB_MIN_C_PARAM:
      min_param_c = get_float(optarg);
      break;
    case O_VB_VERBOSE:
      mode_verbose = 1;
      break;
    /*
    case O_VB_BDEU:
      mode_bdeu = 1;
      bdeu_min_smooth = get_float(optarg);
      break;
    */
    case O_VB_SMOOTH_A:
      num_smooth_w = get_smooth_w(optarg);
      break;
    case O_VB_OCC_ORDER:
      mode_lex_order = 0;
      break;
    case O_VB_RANK:
      mode_rank = 1;
      get_rank(optarg);
      break;
    case O_VB_RANK_HIDE:
      mode_rank_hide = 1;
      break;
    case O_VB_IGNORE:
      mode_ignore = 1;
      add_attr_ignore(optarg);
      break;
    case O_VB_HIDE_ATTR:
      mode_hide_attr = 1;
      break;
    case O_VB_PREC_OUT:
      prec_out = get_int(optarg);
      break;
    case O_VB_PREC_MSG:
      prec_msg = get_int(optarg);
      break;
    case O_VB_TAU:
      tau = get_smooth_tau(optarg);
      break;
    case O_VB_UNCOMPRESS:
      mode_uncompress = 1;
      break;
    case O_VB_NOOP:
      print_msg("Warning: %s not supported\n",optstr);
      /* do nothing */
      break;
#if 0
    case NOT_AVAIL:
      mode_rank_sm = 1;
      num_rank_sm = get_int(optarg);
      break;
#endif
    default:
      break;
    }
  }
}

static void postproc_args(void)
{
  int v,u;

  check_tuple_size();

  /* tuple_size_d and mode_nominal are computed_h here */
  build_attr_info();

  if (mode_nominal) build_smooth_nominal(num_smooth_w);
  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      if (alpha_w < 0 || (alpha_w > 0 && alpha_w < gauss_attr_size[u]))
	die("Missing or illegal value for --alpha (-O) option\n");
    }

    if (smooth_s < 0.0) {
      die("Illegal value for --smooth-var (-L) option\n");
    }
    else if (smooth_s < DEFAULT_SMOOTH_S) {
      smooth_s = DEFAULT_SMOOTH_S;
    }

    if (tau < 0.0) {
      die("Illegal value for --tau (-T) option\n");
    }
    else if (tau < DEFAULT_TAU) {
      tau = DEFAULT_TAU;
    }
  }

  if (mode_id && (attr_id < 0 || attr_id >= tuple_size))
    die("Illegal value for --id (-I) option\n");

  if (mode_eval && (attr_eval < 0 || attr_eval >= tuple_size))
    die("Illegal value for --eval (-E) option\n");

  if (mode_id && mode_eval && attr_id == attr_eval)
    die("Confliction between --id (-I) and --eval (-E) options\n");

  if (mode_ignore) {
    num_attr_ignore = 0;
    for (v=0; v<tuple_size; v++) {
      if (attr_ignore[v]) {
	if (v == attr_id)
	  die("Confliction between --id (-I) and --ignore (-y) options\n");
	else if (v == attr_eval)
	  die("Confliction between --ignore (-y) and --eval (-E) options\n");
	else
	  num_attr_ignore++;
      }
    }
  }

  if (mode_em && mode_search_num_class) {
    if (kmin <= 1 || kmax <= 0 || kstep <= 0 || kmin > kmax) {
      fprintf(stderr,"Illegal value for --n-class (-k) option (Use -h for usage)\n");
      exit(1);
    }
  }
  else if (mode_em && num_class <= 1) {
    fprintf(stderr,"Missing or illegal --n-class (-k) option (Use -h for usage)\n");
    exit(1);
  }
  
  /* In the later versions, this will be more complicated */
  mode_as_is = (mode_em) ? 1 : 0;

  if (max_iterate < 0 ||
      max_init_iterate < 0) {
    fprintf(stderr,"Illegal value for the number of iterations\n");
    exit(1);
  }

  if (max_iterate > MAX_MAX_ITERATE ||
      max_iterate > MAX_MAX_ITERATE) {
    fprintf(stderr,"Maximum number of iterations cannot exceed %d\n",
	    MAX_MAX_ITERATE);
    exit(1);
  }

  if (num_init <= 0) {
    fprintf(stderr,"Illegal value for the number of initial parameter sets\n");
    exit(1);
  }

  if (max_jsplit < 0) {
    fprintf(stderr,"Illegal value for the number of split candidate to be examined\n");
    exit(1);
  }

  if (min_param_c < 0) {
    fprintf(stderr,
	    "Illegal value for the lower limit of prob. "
	    "of the class to be split/merge\n");
    exit(1);
  }

  if (min_new_param_c <= 0.0) {
    fprintf(stderr,
	    "Illegal value for the lower limit of updated prob. "
	    "of the splited/merged class\n");
    exit(1);
  }

#ifdef USE_ICONV
  mode_conv_in = (strcmp(charcode_in,CHARCODE_INTERNAL) == 0) ? 0 : 1;
  mode_conv_out = (strcmp(charcode_out,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#else
  mode_conv_in = 0;
  mode_conv_out = 0;
#endif

  if (mode_bdeu && bdeu_min_smooth <= 0) {
    fprintf(stderr,"Minimum pseudo count for BDeu should be positive (Use -h for usage)\n");
    exit(1);
  }

  if (mode_cluster) {
    switch (cl_type) {
    case CL_MP:
      strncpy(cl_type_str,"MEMBERSHIP_DIST",ENUM_TYPE_LEN);
      break;
    default:
      fprintf(stderr,"Internal error: Illegal Info. Criterion\n");
      exit(1);
    }
  }

  switch (rel_m) {
  case REL_MP:
    strncpy(rel_m_str,"MEMBERSHIP_DIST",ENUM_TYPE_LEN);
    break;
  default:
    fprintf(stderr,"Internal error: Illegal relevance measure between class and word\n");
    exit(1);
  }

  if (min_smem_improve < 0) {
    min_smem_improve = epsilon * MIN_SMEM_IMPROVE_RATIO;
  }
}

int main(int argc, char *argv[])
{
  int v;

  initialize_mode();
  initialize_buffer();
  initialize_desc();
#ifdef USE_ICONV  
  initialize_iconv_sys();
#endif

  initialize_matrix_pool();
  initialize_vector_pool();

  parse_args(argc,argv);
  postproc_args();

#ifdef USE_ICONV
  initialize_iconv();
#endif
  initialize_rand();

  print_settings(fp_log);
  print_model(fp_log);

  if (mode_em)
    vbem_main();
  else
    no_vbem_main();

  fclose(fp_log);
#ifdef USE_ICONV
  iconv_close(cd_in);
  iconv_close(cd_out);
#endif

  for (v=0; v<tuple_size_d; v++) {
    release_hashtable(&ht[v]);
  }

  release_buffer();
  release_desc();

  free(smooth_w); smooth_w = NULL;
  free(mode_smooth_auto_w); mode_smooth_auto_w = NULL;

  return 0;
}
